@@ -1,23 +1,30 @@
-# Faraday Penetration Test IDE
-# Copyright (C) 2016  Infobyte LLC (http://www.infobytesec.com/)
-# See the file 'doc/LICENSE' for the license information
-from io import StringIO
+"""
+Faraday Penetration Test IDE
+Copyright (C) 2016  Infobyte LLC (https://faradaysec.com/)
+See the file 'doc/LICENSE' for the license information
+"""
 
+# Standard library imports
 import logging
 import csv
-import flask
 import re
-from flask import Blueprint, make_response, jsonify, abort
+from io import StringIO
+
+# Related third party imports
+import wtforms
 import pytz
+import flask
+from flask import Blueprint, make_response, jsonify, abort
 from flask_classful import route
+from flask_wtf.csrf import validate_csrf
 from marshmallow import fields, Schema
 from filteralchemy import Filter, FilterSet, operators
 from sqlalchemy import desc
-import wtforms
-from flask_wtf.csrf import validate_csrf
+from sqlalchemy.orm import joinedload, undefer
 
+# Local application imports
 from faraday.server.utils.database import get_or_create
-
+from faraday.server.utils.command import set_command_id
 from faraday.server.api.base import (
     ReadWriteWorkspacedView,
     PaginatedMixin,
@@ -25,40 +32,38 @@ from faraday.server.api.base import (
     FilterAlchemyMixin,
     FilterSetMeta,
     FilterWorkspacedMixin,
-    PatchableWorkspacedMixin
+    BulkDeleteWorkspacedMixin,
+    BulkUpdateWorkspacedMixin, CountMultiWorkspacedMixin, get_workspace,
 )
+from faraday.server.api.modules.services import ServiceSchema
 from faraday.server.schemas import (
     MetadataSchema,
     MutableField,
     NullToBlankString,
     PrimaryKeyRelatedField,
-    SelfNestedField
+    SelfNestedField,
 )
 from faraday.server.models import Host, Service, db, Hostname, CommandObject, Command
-from faraday.server.api.modules.services import ServiceSchema
-
 host_api = Blueprint('host_api', __name__)
-
 logger = logging.getLogger(__name__)
 
 
+def get_total_count(obj):
+    return obj.vulnerability_critical_generic_count + obj.vulnerability_high_generic_count \
+           + obj.vulnerability_medium_generic_count + obj.vulnerability_low_generic_count \
+           + obj.vulnerability_info_generic_count + obj.vulnerability_unclassified_generic_count
+
+
 class HostCountSchema(Schema):
-    host_id = fields.Integer(dump_only=True, allow_none=False,
-                                 attribute='id')
-    critical = fields.Integer(dump_only=True, allow_none=False,
-                                 attribute='vulnerability_critical_count')
-    high = fields.Integer(dump_only=True, allow_none=False,
-                              attribute='vulnerability_high_count')
-    med = fields.Integer(dump_only=True, allow_none=False,
-                              attribute='vulnerability_medium_count')
-    low = fields.Integer(dump_only=True, allow_none=False,
-                              attribute='vulnerability_low_count')
-    info = fields.Integer(dump_only=True, allow_none=False,
-                              attribute='vulnerability_informational_count')
+    host_id = fields.Integer(dump_only=True, allow_none=False, attribute='id')
+    critical = fields.Integer(dump_only=True, allow_none=False, attribute='vulnerability_critical_generic_count')
+    high = fields.Integer(dump_only=True, allow_none=False, attribute='vulnerability_high_generic_count')
+    med = fields.Integer(dump_only=True, allow_none=False, attribute='vulnerability_medium_generic_count')
+    low = fields.Integer(dump_only=True, allow_none=False, attribute='vulnerability_low_generic_count')
+    info = fields.Integer(dump_only=True, allow_none=False, attribute='vulnerability_info_generic_count')
     unclassified = fields.Integer(dump_only=True, allow_none=False,
-                              attribute='vulnerability_unclassified_count')
-    total = fields.Integer(dump_only=True, allow_none=False,
-                                 attribute='vulnerability_total_count')
+                                  attribute='vulnerability_unclassified_generic_count')
+    total = fields.Function(get_total_count, dump_only=True)
 
 
 class HostSchema(AutoSchema):
@@ -74,7 +79,6 @@ class HostSchema(AutoSchema):
     owned = fields.Boolean(default=False)
     owner = PrimaryKeyRelatedField('username', attribute='creator', dump_only=True)
     services = fields.Integer(attribute='open_service_count', dump_only=True)
-    vulns = fields.Integer(attribute='vulnerability_count', dump_only=True)
     credentials = fields.Integer(attribute='credentials_count', dump_only=True)
     hostnames = MutableField(
         PrimaryKeyRelatedField('name', many=True,
@@ -86,8 +90,10 @@ class HostSchema(AutoSchema):
     type = fields.Function(lambda obj: 'Host', dump_only=True)
     service_summaries = fields.Method('get_service_summaries', dump_only=True)
     versions = fields.Method('get_service_version', dump_only=True)
-    important = fields.Boolean(default=False)
+    importance = fields.Integer(default=0, validate=lambda stars: stars in [0, 1, 2, 3])
     severity_counts = SelfNestedField(HostCountSchema(), dump_only=True)
+    command_id = fields.Int(required=False, load_only=True)
+    vulns = fields.Function(get_total_count, dump_only=True)
 
     class Meta:
         model = Host
@@ -95,15 +101,17 @@ class HostSchema(AutoSchema):
                   'credentials', 'default_gateway', 'metadata',
                   'name', 'os', 'owned', 'owner', 'services', 'vulns',
                   'hostnames', 'type', 'service_summaries', 'versions',
-                  'important', 'severity_counts'
+                  'importance', 'severity_counts', 'command_id'
                   )
 
-    def get_service_summaries(self, obj):
+    @staticmethod
+    def get_service_summaries(obj):
         return [service.summary
                 for service in obj.services
                 if service.status == 'open']
 
-    def get_service_version(self, obj):
+    @staticmethod
+    def get_service_version(obj):
         return [service.version
                 for service in obj.services
                 if service.status == 'open']
@@ -135,29 +143,81 @@ class HostFilterSet(FilterSet):
     port = ServicePortFilter(fields.Str())
 
 
-
 class HostsView(PaginatedMixin,
                 FilterAlchemyMixin,
                 ReadWriteWorkspacedView,
-                FilterWorkspacedMixin):
+                FilterWorkspacedMixin,
+                CountMultiWorkspacedMixin,
+                BulkDeleteWorkspacedMixin,
+                BulkUpdateWorkspacedMixin):
     route_base = 'hosts'
     model_class = Host
-    order_field = desc(Host.vulnerability_critical_generic_count),\
-        desc(Host.vulnerability_high_generic_count),\
-        desc(Host.vulnerability_medium_generic_count),\
-        desc(Host.vulnerability_low_generic_count),\
-        desc(Host.vulnerability_info_generic_count),\
+    order_field = desc(Host.vulnerability_critical_generic_count), \
+        desc(Host.vulnerability_high_generic_count), \
+        desc(Host.vulnerability_medium_generic_count), \
+        desc(Host.vulnerability_low_generic_count), \
+        desc(Host.vulnerability_info_generic_count), \
         desc(Host.vulnerability_unclassified_generic_count), Host.ip.asc()
 
     schema_class = HostSchema
     filterset_class = HostFilterSet
     get_undefer = [Host.credentials_count,
                    Host.open_service_count,
-                   Host.vulnerability_count]
+                   Host.vulnerability_critical_generic_count,
+                   Host.vulnerability_high_generic_count,
+                   Host.vulnerability_medium_generic_count,
+                   Host.vulnerability_low_generic_count,
+                   Host.vulnerability_info_generic_count,
+                   Host.vulnerability_unclassified_generic_count,
+                   ]
     get_joinedloads = [Host.hostnames, Host.services, Host.update_user]
 
-    def _get_base_query(self, workspace_name):
-        return Host.query_with_count(None, None, workspace_name)
+    def _get_eagerloaded_query(self, *args, **kwargs):
+        """
+        Overrides _get_eagerloaded_query of GenericView
+        """
+        options = []
+        try:
+            has_creator = 'owner' in self._get_schema_class().opts.fields
+        except AttributeError:
+            has_creator = False
+        show_stats = kwargs.pop('show_stats', True)
+        if has_creator:
+            # APIs for objects with metadata always return the creator's
+            # username. Do a joinedload to prevent doing one query per object
+            # (n+1) problem
+            options.append(joinedload(
+                getattr(self.model_class, 'creator')).load_only('username'))
+        query = self._get_base_query(*args, **kwargs)
+        options += [joinedload(relationship)
+                    for relationship in self.get_joinedloads]
+        if show_stats:
+            options += [undefer(column) for column in self.get_undefer]
+        return query.options(*options)
+
+    def index(self, **kwargs):
+        """
+          ---
+          get:
+            summary: "Get a list of hosts."
+            tags: ["Host"]
+            responses:
+              200:
+                description: Ok
+                content:
+                  application/json:
+                    schema: HostSchema
+          tags: ["Host"]
+          responses:
+            200:
+              description: Ok
+        """
+        kwargs['show_stats'] = flask.request.args.get('stats', '') != 'false'
+
+        if not kwargs['show_stats']:
+            kwargs['exclude'] = ['severity_counts', 'vulns', 'credentials', 'services']
+
+        return super().index(**kwargs)
 
     @route('/filter')
     def filter(self, workspace_name):
@@ -193,7 +253,7 @@ class HostsView(PaginatedMixin,
         pagination_metadata.total = count
         return self._envelope_list(filtered_objs, pagination_metadata)
 
-    @route('/bulk_create/', methods=['POST'])
+    @route('/bulk_create', methods=['POST'])
     def bulk_create(self, workspace_name):
         """
         ---
@@ -224,7 +284,7 @@ class HostsView(PaginatedMixin,
             items = re.findall(r"([.a-zA-Z0-9_-]+)", list_string)
             return items
 
-        workspace = self._get_workspace(workspace_name)
+        workspace = get_workspace(workspace_name)
 
         logger.info("Create hosts from CSV")
         if 'file' not in flask.request.files:
@@ -242,9 +302,9 @@ class HostsView(PaginatedMixin,
             for host_dict in hosts_reader:
                 try:
                     hostnames = parse_hosts(host_dict.pop('hostnames'))
-                    other_fields = {'owned': False, 'mac': u'00:00:00:00:00:00', 'default_gateway_ip': u'None'}
+                    other_fields = {'owned': False, 'mac': '00:00:00:00:00:00', 'default_gateway_ip': 'None'}
                     host_dict.update(other_fields)
-                    host = super(HostsView, self)._perform_create(host_dict, workspace_name)
+                    host = super()._perform_create(host_dict, workspace_name)
                     host.workspace = workspace
                     for name in hostnames:
                         get_or_create(db.session, Hostname, name=name, host=host, workspace=host.workspace)
@@ -255,12 +315,14 @@ class HostsView(PaginatedMixin,
                 else:
                     logger.debug("Host Created (%s)", host_dict)
                     hosts_created_count += 1
-            return make_response(jsonify(hosts_created=hosts_created_count, hosts_with_errors=hosts_with_errors_count), 200)
+            logger.info("Hosts created in bulk")
+            return make_response(jsonify(hosts_created=hosts_created_count,
+                                         hosts_with_errors=hosts_with_errors_count), 200)
         except Exception as e:
             logger.error("Error parsing hosts CSV (%s)", e)
             abort(400, f"Error parsing hosts CSV ({e})")
 
-    @route('/<host_id>/services/')
+    @route('/<host_id>/services')
     def service_list(self, workspace_name, host_id):
         """
         ---
@@ -281,7 +343,7 @@ class HostsView(PaginatedMixin,
         services = self._get_object(host_id, workspace_name).services
         return ServiceSchema(many=True).dump(services)
 
-    @route('/countVulns/')
+    @route('/countVulns')
     def count_vulns(self, workspace_name):
         """
         ---
@@ -299,6 +361,8 @@ class HostsView(PaginatedMixin,
           200:
             description: Ok
         """
+        workspace = get_workspace(workspace_name)
+
         host_ids = flask.request.args.get('hosts', None)
         if host_ids:
             host_id_list = host_ids.split(',')
@@ -308,15 +372,14 @@ class HostsView(PaginatedMixin,
         res_dict = {'hosts': {}}
 
         host_count_schema = HostCountSchema()
-        host_count = Host.query_with_count(None, host_id_list, workspace_name)
+        host_count = Host.query_with_count(host_id_list, workspace)
 
         for host in host_count.all():
             res_dict["hosts"][host.id] = host_count_schema.dump(host)
-        # return counts.data
 
         return res_dict
 
-    @route('/<host_id>/tools_history/')
+    @route('/<host_id>/tools_history')
     def tool_impacted_by_host(self, workspace_name, host_id):
         """
         ---
@@ -334,7 +397,7 @@ class HostsView(PaginatedMixin,
           200:
             description: Ok
         """
-        workspace = self._get_workspace(workspace_name)
+        workspace = get_workspace(workspace_name)
         query = db.session.query(Host, Command).filter(Host.id == CommandObject.object_id,
                                                        CommandObject.object_type == 'host',
                                                        Command.id == CommandObject.command_id,
@@ -344,15 +407,22 @@ class HostsView(PaginatedMixin,
         res_dict = {'tools': []}
         for row in result:
             _, command = row
-            res_dict['tools'].append({'command': command.tool, 'user': command.user, 'params': command.params, 'command_id': command.id, 'create_date': command.create_date.replace(tzinfo=pytz.utc).strftime("%c")})
+            res_dict['tools'].append({'command': command.tool,
+                                      'user': command.user,
+                                      'params': command.params,
+                                      'command_id': command.id,
+                                      'create_date': command.create_date.replace(tzinfo=pytz.utc).isoformat()})
         return res_dict
 
     def _perform_create(self, data, **kwargs):
         hostnames = data.pop('hostnames', [])
-        host = super(HostsView, self)._perform_create(data, **kwargs)
+        command_id = data.pop('command_id', None)
+        host = super()._perform_create(data, **kwargs)
         for name in hostnames:
             get_or_create(db.session, Hostname, name=name, host=host,
                           workspace=host.workspace)
+        if command_id:
+            set_command_id(db.session, host, True, command_id)
         db.session.commit()
         return host
 
@@ -367,10 +437,10 @@ class HostsView(PaginatedMixin,
         # A commit is required here, otherwise it breaks (i'm not sure why)
         db.session.commit()
 
-        return super(HostsView, self)._update_object(obj, data)
+        return super()._update_object(obj, data)
 
     def _filter_query(self, query):
-        query = super(HostsView, self)._filter_query(query)
+        query = super()._filter_query(query)
         search_term = flask.request.args.get('search', None)
         if search_term is not None:
             like_term = '%' + search_term + '%'
@@ -379,12 +449,48 @@ class HostsView(PaginatedMixin,
                 Service.name.ilike(like_term))
             match_os = Host.os.ilike(like_term)
             match_hostname = Host.hostnames.any(Hostname.name.ilike(like_term))
-            query = query.filter(match_ip |
-                                 match_service_name |
-                                 match_os |
-                                 match_hostname)
+            query = query.filter(match_ip
+                                 | match_service_name
+                                 | match_os
+                                 | match_hostname)
         return query
 
+    def patch(self, object_id, workspace_name=None, **kwargs):
+        """
+        ---
+          tags: ["{tag_name}"]
+          summary: Updates {class_model}
+          parameters:
+          - in: path
+            name: object_id
+            required: true
+            schema:
+              type: integer
+          - in: path
+            name: workspace_name
+            required: true
+            schema:
+              type: string
+          requestBody:
+            required: true
+            content:
+              application/json:
+                schema: {schema_class}
+          responses:
+            200:
+              description: Ok
+              content:
+                application/json:
+                  schema: {schema_class}
+            409:
+              description: Duplicated key found
+              content:
+                application/json:
+                  schema: {schema_class}
+        """
+        kwargs['exclude'] = ['severity_counts']
+        return super().patch(object_id, workspace_name=workspace_name, **kwargs)
+
     def _envelope_list(self, objects, pagination_metadata=None):
         hosts = []
         for index, host in enumerate(objects):
@@ -397,73 +503,27 @@ class HostsView(PaginatedMixin,
             })
         return {
             'rows': hosts,
-            'count': (pagination_metadata and pagination_metadata.total
-                           or len(hosts)),
+            'count': (pagination_metadata and pagination_metadata.total or len(hosts)),
         }
 
-    # TODO SCHEMA
-    @route('bulk_delete/', methods=['DELETE'])
-    def bulk_delete(self, workspace_name):
-        """
-        ---
-        delete:
-          tags: ["Bulk", "Host"]
-          description: Delete hosts in bulk
-          responses:
-            200:
-              description: Ok
-            400:
-              description: Bad request
-            403:
-              description: Forbidden
-        tags: ["Bulk", "Host"]
-        responses:
-          200:
-            description: Ok
-        """
-        workspace = self._get_workspace(workspace_name)
-        json_request = flask.request.get_json()
-        if not json_request:
-            flask.abort(400, 'Invalid request. Check the request data or the content type of the request')
-        hosts_ids = json_request.get('hosts_ids', [])
-        hosts_ids = [host_id for host_id in hosts_ids if isinstance(host_id, int)]
-        deleted_hosts = 0
-        if hosts_ids:
-            deleted_hosts = Host.query.filter(
-                Host.id.in_(hosts_ids),
-                Host.workspace_id == workspace.id).delete(synchronize_session='fetch')
-        else:
-            flask.abort(400, "Invalid request")
+    @route('', methods=['DELETE'])
+    def bulk_delete(self, workspace_name, **kwargs):
+        # TODO REVISE ORIGINAL METHOD TO UPDATE NEW METHOD
+        return BulkDeleteWorkspacedMixin.bulk_delete(self, workspace_name, **kwargs)
 
-        db.session.commit()
-        response = {'deleted_hosts': deleted_hosts}
-        return flask.jsonify(response)
+    bulk_delete.__doc__ = BulkDeleteWorkspacedMixin.bulk_delete.__doc__
 
+    def _pre_bulk_update(self, data, **kwargs):
+        hostnames = data.pop('hostnames', None)
+        ans_data = super()._pre_bulk_update(data, **kwargs)
+        if hostnames is not None:
+            ans_data["hostnames"] = hostnames
+        return ans_data
 
-class HostsV3View(HostsView, PatchableWorkspacedMixin):
-    route_prefix = '/v3/ws/<workspace_name>/'
-    trailing_slash = False
+    def _post_bulk_update(self, ids, extracted_data, **kwargs):
+        if "hostnames" in extracted_data:
+            for obj in self._bulk_update_query(ids, **kwargs).all():
+                obj.set_hostnames(extracted_data["hostnames"])
 
-    @route('/<host_id>/services')
-    def service_list(self, workspace_name, host_id):
-        return super(HostsV3View, self).service_list(workspace_name, host_id)
-
-    @route('/<host_id>/tools_history')
-    def tool_impacted_by_host(self, workspace_name, host_id):
-        return super(HostsV3View, self).tool_impacted_by_host(workspace_name, host_id)
-
-    @route('/bulk_create', methods=['POST'])
-    def bulk_create(self, workspace_name):
-        return super(HostsV3View, self).bulk_create(workspace_name)
-
-    @route('/countVulns')
-    def count_vulns(self, workspace_name):
-        return super(HostsV3View, self).count_vulns()
-
-    service_list.__doc__ = HostsView.service_list.__doc__
-    tool_impacted_by_host.__doc__ = HostsView.tool_impacted_by_host.__doc__
-    bulk_create.__doc__ = HostsView.bulk_create.__doc__
-    count_vulns.__doc__ = HostsView.count_vulns.__doc__
 
 HostsView.register(host_api)
-HostsV3View.register(host_api)
