@@ -14,20 +14,56 @@
     :license: GNU AGPLv3+ or BSD
 
 """
+# Standard library imports
 import inspect
 import logging
+from datetime import date
 
-from sqlalchemy import func
-from sqlalchemy import and_, or_
-from sqlalchemy import inspect as sqlalchemy_inspect
+# Related third party imports
+from sqlalchemy import (
+    and_,
+    or_,
+    func,
+    nullsfirst,
+    nullslast,
+    inspect as sqlalchemy_inspect, text,
+)
 from sqlalchemy.ext.associationproxy import AssociationProxy
-from sqlalchemy.orm.attributes import InstrumentedAttribute
-from sqlalchemy.orm.attributes import QueryableAttribute
+from sqlalchemy.ext.hybrid import hybrid_property
 from sqlalchemy.orm import ColumnProperty
+from sqlalchemy.orm.attributes import InstrumentedAttribute, QueryableAttribute
 
+# Local application imports
+from faraday.server.models import (
+    User,
+    CVE,
+    Role,
+    CustomFieldsSchema,
+    VulnerabilityGeneric,
+    Vulnerability,
+    VulnerabilityWeb,
+)
+
+VULNERABILITY_MODELS = (Vulnerability, VulnerabilityWeb, VulnerabilityGeneric)
 
 logger = logging.getLogger(__name__)
 
+bind_counter = 0
+
+
+def get_bind_counter():
+    return bind_counter
+
+
+def increment_bind_counter():
+    global bind_counter
+    bind_counter += 1
+
+
+def reset_bind_counter():
+    global bind_counter
+    bind_counter = 0
+
 
 def session_query(session, model):
     """Returns a SQLAlchemy query object for the specified `model`.
@@ -74,6 +110,7 @@ def _sub_operator(model, argument, fieldname):
     This function is for use with the ``has`` and ``any`` search operations.
 
     """
+    submodel = None
     if isinstance(model, InstrumentedAttribute):
         submodel = model.property.mapper.class_
     elif isinstance(model, AssociationProxy):
@@ -87,8 +124,7 @@ def _sub_operator(model, argument, fieldname):
         relation = None
         if '__' in fieldname:
             fieldname, relation = fieldname.split('__')
-        return QueryBuilder._create_operation(submodel, fieldname, operator,
-                                              argument, relation)
+        return QueryBuilder._create_operation(submodel, fieldname, operator, argument, relation)
     # Support legacy has/any with implicit eq operator
     return getattr(submodel, fieldname) == argument
 
@@ -112,8 +148,8 @@ def _sub_operator(model, argument, fieldname):
 #: be described by the strings ``'=='``, ``'eq'``, ``'equals'``, etc.
 OPERATORS = {
     # Operators which accept a single argument.
-    'is_null': lambda f: f == None,
-    'is_not_null': lambda f: f != None,
+    'is_null': lambda f: f == None,  # noqa E711
+    'is_not_null': lambda f: f != None,  # noqa E711
     'desc': lambda f: f.desc,
     'asc': lambda f: f.asc,
     # Operators which accept two arguments.
@@ -145,9 +181,63 @@ OPERATORS = {
     # Operators which accept three arguments.
     'has': lambda f, a, fn: f.has(_sub_operator(f, a, fn)),
     'any': lambda f, a, fn: f.any(_sub_operator(f, a, fn)),
+    'not_any': lambda f, a, fn: ~f.any(_sub_operator(f, a, fn)),
+    # Custom operator for json fields
+    'json': lambda f: f,
 }
 
 
+def get_json_operator(operator):
+    operator_mapping = {
+        '==': ('=', 'compare'),
+        'eq': ('=', 'compare'),
+        'equals': ('=', 'compare'),
+        'equal_to': ('=', 'compare'),
+        '!=': ('<>', 'compare'),
+        'ne': ('<>', 'compare'),
+        'neq': ('<>', 'compare'),
+        'not_equal_to': ('<>', 'compare'),
+        'does_not_equal': ('<>', 'compare'),
+        '<': ('<', 'compare'),
+        'lt': ('<', 'compare'),
+        '<=': ('<=', 'compare'),
+        'le': ('<=', 'compare'),
+        'lte': ('<=', 'compare'),
+        'leq': ('<=', 'compare'),
+        '>': ('>', 'compare'),
+        'gt': ('>', 'compare'),
+        '>=': ('>=', 'compare'),
+        'ge': ('>=', 'compare'),
+        'gte': ('>=', 'compare'),
+        'geq': ('>=', 'compare'),
+        'like': ('LIKE', 'compare'),
+        'ilike': ('ILIKE', 'compare'),
+        'any': ('@>', 'any'),
+        'not_any': ('@>', 'not_any'),
+        'is_null': ('IS NULL', 'exists'),
+        'is_not_null': ('IS NOT NULL', 'exists'),
+    }
+
+    return operator_mapping.get(operator, None)
+
+
+def get_json_query(table, field, op, op_type, counter):
+    if op_type == 'compare':
+        return f"{table}.{field} ->> :key_{counter} {op} :value_{counter}"  # nosec
+    elif op_type == 'exists':
+        return f"{table}.{field} ->> :key_{counter} {op}"  # nosec
+    elif op_type == 'any':
+        return f"{table}.{field} -> :key_{counter} {op} :value_{counter}"  # nosec
+    elif op_type == 'not_any':
+        return f"NOT {table}.{field} -> :key_{counter} {op} :value_{counter} OR {table}.{field} -> :key_{counter} IS NULL"  # nosec
+    elif op_type == 'list_contains':
+        return f"EXISTS (SELECT 1 FROM jsonb_array_elements_text({table}.{field} -> :key_{counter}) AS element WHERE element {op} :value_{counter})"  # nosec
+    elif op_type == 'date':
+        return f"({table}.{field} ->> :key_{counter})::DATE {op} :value_{counter}"  # nosec
+    else:
+        raise TypeError('Invalid filters')
+
+
 class OrderBy:
     """Represents an "order by" in a SQL query expression."""
 
@@ -222,8 +312,7 @@ class Filter:
 
     def __repr__(self):
         """Returns a string representation of this object."""
-        return '<Filter {0} {1} {2}>'.format(self.fieldname, self.operator,
-                                             self.argument or self.otherfield)
+        return f'<Filter {self.fieldname} {self.operator} {self.argument or self.otherfield}>'
 
     @staticmethod
     def from_dictionary(dictionary):
@@ -280,6 +369,7 @@ class Filter:
 class JunctionFilter(Filter):
     def __init__(self, *subfilters):
         self.subfilters = subfilters
+
     def __iter__(self):
         return iter(self.subfilters)
 
@@ -387,7 +477,6 @@ class QueryBuilder:
     a given model.
 
     """
-
     @staticmethod
     def _create_operation(model, fieldname, operator, argument, relation=None):
         """Translates an operation described as a string to a valid SQLAlchemy
@@ -411,7 +500,7 @@ class QueryBuilder:
         operation will be applied to the field with name `fieldname` on the
         entity related to `model` whose name, as a string, is `relation`.
 
-        `operation` is a string representating the operation which will be
+        `operation` is a string representing the operation which will be
          executed between the field and the argument received. For example,
          ``'gt'``, ``'lt'``, ``'like'``, ``'in'`` etc.
 
@@ -431,30 +520,65 @@ class QueryBuilder:
           `relation` exists on `model`
 
         """
-        # raises KeyError if operator not in OPERATORS
-        opfunc = OPERATORS[operator]
-        # In Python 3.0 or later, this should be `inspect.getfullargspec`
-        # because `inspect.getargspec` is deprecated.
-        numargs = len(inspect.getargspec(opfunc).args)
-        # raises AttributeError if `fieldname` or `relation` does not exist
-        field = getattr(model, relation or fieldname)
-        # each of these will raise a TypeError if the wrong number of argments
-        # is supplied to `opfunc`.
-        if numargs == 1:
-            return opfunc(field)
-        if argument is None:
-            msg = ('To compare a value to NULL, use the is_null/is_not_null '
-                   'operators.')
-            raise TypeError(msg)
-        if numargs == 2:
-            map_attr = {
-                'creator': 'username',
-            }
-            if hasattr(getattr(field, 'prop'), 'entity'):
-                field = getattr(field.comparator.entity.class_, map_attr.get(field.prop.key, field.prop.key))
+        if '->' in fieldname:
+            if getattr(model, fieldname.split('->')[0]):
+                table = 'vulnerability' if model in VULNERABILITY_MODELS else model.__tablename__
 
-            return opfunc(field, argument)
-        return opfunc(field, argument, fieldname)
+                field, key = fieldname.split('->')
+                custom_field = CustomFieldsSchema.query.filter(CustomFieldsSchema.field_name == key).first()
+
+                try:
+                    op, op_type = get_json_operator(operator)
+                except TypeError as e:
+                    raise TypeError('Invalid filters') from e
+
+                if op in ['like', 'ilike']:
+                    if custom_field.field_type == 'list':
+                        op_type = 'list_contains'
+
+                if op_type in ['any', 'not_any']:
+                    if not argument.isnumeric():
+                        argument = f"\"{argument}\""
+
+                if custom_field.field_type == 'date':
+                    argument = date.fromisoformat(argument)
+                    op_type = 'date'
+
+                increment_bind_counter()
+
+                bindparams = {
+                    f'key_{get_bind_counter()}': key,
+                    f'value_{get_bind_counter()}': argument,
+                }
+
+                query = get_json_query(table, field, op, op_type, get_bind_counter())
+
+                return OPERATORS['json'](text(f"{query}").bindparams(**bindparams))
+        else:
+            # raises KeyError if operator not in OPERATORS
+            opfunc = OPERATORS[operator]
+            # In Python 3.0 or later, this should be `inspect.getfullargspec`
+            # because `inspect.getargspec` is deprecated.
+            numargs = len(inspect.getargspec(opfunc).args)
+            # raises AttributeError if `fieldname` or `relation` does not exist
+            field = getattr(model, relation or fieldname)
+            # each of these will raise a TypeError if the wrong number of arguments
+            # is supplied to `opfunc`.
+            if numargs == 1:
+                return opfunc(field)
+            if argument is None:
+                msg = ('To compare a value to NULL, use the is_null/is_not_null '
+                       'operators.')
+                raise TypeError(msg)
+            if numargs == 2:
+                map_attr = {
+                    'creator': 'username',
+                }
+                if hasattr(field, 'prop') and hasattr(getattr(field, 'prop'), 'entity'):
+                    field = getattr(field.comparator.entity.class_, map_attr.get(field.prop.key, field.prop.key))
+
+                return opfunc(field, argument)
+            return opfunc(field, argument, fieldname)
 
     @staticmethod
     def _create_filter(model, filt):
@@ -494,13 +618,20 @@ class QueryBuilder:
         create_filt = QueryBuilder._create_filter
 
         def create_filters(filt):
-            if not getattr(filt, 'fieldname', False) or filt.fieldname.split('__')[0] in valid_model_fields:
-                try:
-                    return create_filt(model, filt)
-                except AttributeError:
-                    # Can't create the filter since the model or submodel does not have the attribute (usually mapper)
-                    return None
-            return None
+            if not isinstance(filt, JunctionFilter) and '__' in filt.fieldname:
+                return create_filt(model, filt)
+            else:
+                if (not getattr(filt, 'fieldname', False)
+                        or filt.fieldname.split('__')[0] in valid_model_fields
+                        or filt.fieldname.split('->')[0] in valid_model_fields):
+                    try:
+                        return create_filt(model, filt)
+                    except AttributeError as e:
+                        # Can't create the filter since the model or submodel
+                        # does not have the attribute (usually mapper)
+                        raise AttributeError(f"Foreign field {filt.fieldname.split('__')[0]} "
+                                             f"not found in submodel") from e
+            raise AttributeError(f"Field {filt.fieldname} not found in model")
 
         return create_filters
 
@@ -534,21 +665,46 @@ class QueryBuilder:
         documentation for :func:`_create_operation` for more information.
 
         """
+        # TODO: Can't this be done with group by below?
+        joined_models = set()
+        query = session.query(model)
+
         if search_params.group_by:
             select_fields = [func.count()]
-            for groupby in search_params.group_by:
-                select_fields.append(getattr(model, groupby.field))
+            for group_by in search_params.group_by:
+                field_name = group_by.field
+                if '__' in field_name:
+                    field_name, field_name_in_relation = field_name.split('__')
+                    relation = getattr(model, field_name)
+                    relation_model = relation.mapper.class_
+                    field = getattr(relation_model, field_name_in_relation)
+                    if relation_model not in joined_models:
+                        if relation_model == User:
+                            query = query.join(relation_model, model.creator_id == relation_model.id)
+                        elif relation_model == CVE:
+                            query = query.join(relation_model, model.cve_instances)
+                        else:
+                            query = query.join(relation_model)
+                    joined_models.add(relation_model)
+                    select_fields.append(field)
+                else:
+                    select_fields.append(getattr(model, group_by.field))
+                query = query.with_entities(*select_fields)
 
-            query = session.query(*select_fields)
-        else:
-            query = session.query(model)
         # This function call may raise an exception.
-        valid_model_fields = [str(algo).split('.')[1] for algo in sqlalchemy_inspect(model).attrs]
+        valid_model_fields = []
+        for orm_descriptor in sqlalchemy_inspect(model).all_orm_descriptors:
+            if isinstance(orm_descriptor, InstrumentedAttribute):
+                valid_model_fields.append(str(orm_descriptor).split('.')[1])
+            if isinstance(orm_descriptor, hybrid_property):
+                valid_model_fields.append(orm_descriptor.__name__)
+        valid_model_fields += [str(algo).split('.')[1] for algo in sqlalchemy_inspect(model).relationships]
 
         filters_generator = map(   # pylint: disable=W1636
             QueryBuilder.create_filters_func(model, valid_model_fields),
             search_params.filters
         )
+
         filters = [filt for filt in filters_generator if filt is not None]
 
         # Multiple filter criteria at the top level of the provided search
@@ -559,24 +715,36 @@ class QueryBuilder:
         # parameters, order by primary key.
         if not _ignore_order_by:
             if search_params.order_by:
-                joined_models = set()
                 for val in search_params.order_by:
                     field_name = val.field
                     if '__' in field_name:
-                        field_name, field_name_in_relation = \
-                            field_name.split('__')
+                        field_name, field_name_in_relation = field_name.split('__')
                         relation = getattr(model, field_name)
                         relation_model = relation.mapper.class_
+                        if relation_model not in joined_models:
+                            # TODO: is it possible to guess if relationship is a many to many
+                            if relation_model == Role:
+                                query = query.join(relation_model, User.roles)
+                            elif relation_model == User:
+                                query = query.join(relation_model, model.creator_id == relation_model.id)
+                            elif relation_model == CVE:
+                                query = query.join(relation_model, model.cve_instances)
+                            else:
+                                query = query.join(relation_model, isouter=True)
+                        joined_models.add(relation_model)
                         field = getattr(relation_model, field_name_in_relation)
                         direction = getattr(field, val.direction)
-                        if relation_model not in joined_models:
-                            query = query.join(relation_model)
-                        joined_models.add(relation_model)
-                        query = query.order_by(direction())
+                        if val.direction == 'desc':
+                            query = query.order_by(nullslast(direction()))
+                        else:
+                            query = query.order_by(nullsfirst(direction()))
                     else:
                         field = getattr(model, val.field)
                         direction = getattr(field, val.direction)
-                        query = query.order_by(direction())
+                        if val.direction == 'desc':
+                            query = query.order_by(nullslast(direction()))
+                        else:
+                            query = query.order_by(nullsfirst(direction()))
             else:
                 if not search_params.group_by:
                     pks = primary_key_names(model)
@@ -585,8 +753,15 @@ class QueryBuilder:
 
         # Group the query.
         if search_params.group_by:
-            for groupby in search_params.group_by:
-                field = getattr(model, groupby.field)
+            for group_by in search_params.group_by:
+                field_name = group_by.field
+                if '__' in field_name:
+                    field_name, field_name_in_relation = field_name.split('__')
+                    relation = getattr(model, field_name)
+                    relation_model = relation.mapper.class_
+                    field = getattr(relation_model, field_name_in_relation)
+                else:
+                    field = getattr(model, group_by.field)
                 query = query.group_by(field)
 
         # Apply limit and offset to the query.
@@ -598,18 +773,18 @@ class QueryBuilder:
         return query
 
 
-def create_query(session, model, searchparams, _ignore_order_by=False):
+def create_query(session, model, search_params, _ignore_order_by=False):
     """Returns a SQLAlchemy query object on the given `model` where the search
-    for the query is defined by `searchparams`.
+    for the query is defined by `search_params`.
 
     The returned query matches the set of all instances of `model` which meet
-    the parameters of the search given by `searchparams`. For more information
+    the parameters of the search given by `search_params`. For more information
     on search parameters, see :ref:`search`.
 
     `model` is a SQLAlchemy declarative model representing the database model
     to query.
 
-    `searchparams` is either a dictionary (as parsed from a JSON request from
+    `search_params` is either a dictionary (as parsed from a JSON request from
     the client, for example) or a :class:`SearchParameters` instance defining
     the parameters of the query (as returned by
     :func:`SearchParameters.from_dictionary`, for example).
@@ -620,9 +795,9 @@ def create_query(session, model, searchparams, _ignore_order_by=False):
     work around a limitation in SQLAlchemy.)
 
     """
-    if isinstance(searchparams, dict):
-        searchparams = SearchParameters.from_dictionary(searchparams)
-    return QueryBuilder.create_query(session, model, searchparams,
+    if isinstance(search_params, dict):
+        search_params = SearchParameters.from_dictionary(search_params)
+    return QueryBuilder.create_query(session, model, search_params,
                                      _ignore_order_by)
 
 
@@ -665,4 +840,5 @@ def search(session, model, search_params, _ignore_order_by=False):
     if is_single:
         # may raise NoResultFound or MultipleResultsFound
         return query.one()
+    reset_bind_counter()
     return query
